import sympy as sp
from sympy.physics.quantum import TensorProduct
from sympy.physics.quantum.dagger import Dagger

def kron(*matrices):
    output = sp.eye(1)
    for i in range(len(matrices)):
        output = TensorProduct(output, matrices[i])
    return output

def identity(dim):
    if dim == 0:
        return sp.eye(1)
    else:
        return sp.eye(dim)

def adjoint(matrix):
    return Dagger(matrix)


import sympy as sp
from . import sp_func

class QubitGate:
    def __init__(self):
        self.gate_type = None
        self.matrix = None
        self._SWAP = sp.Matrix([
            [1,0,0,0],
            [0,0,1,0],
            [0,1,0,0],
            [0,0,0,1],
        ])

class SingleQubitGate(QubitGate):
    def __init__(self):
        super().__init__()
        self.gate_type = "SingleQubitGate"

class TwoQubitGate(QubitGate):
    def __init__(self):
        super().__init__()
        self.gate_type = "TwoQubitGate"
    def get_submatrix(self, wire1, wire2):
        wire_min = min(wire1, wire2)
        wire_max = max(wire1, wire2)
        if wire_min + 1 == wire_max:
            if wire_min == wire1:
                return self.matrix
            elif wire_min == wire2:
                return self._SWAP * self.matrix * self._SWAP
        else:
            swap_gate = sp_func.kron(sp_func.identity(2**((wire_max-wire_min)-1)), self._SWAP)
            if wire_min == wire1:
                return swap_gate * sp_func.kron(self.get_submatrix(wire_min, wire_max-1), sp_func.identity(2)) * swap_gate
            elif wire_min == wire2:
                return swap_gate * sp_func.kron(self.get_submatrix(wire_max-1, wire_min), sp_func.identity(2)) * swap_gate

class H(SingleQubitGate):
    def __init__(self, wire):
        super().__init__()
        self.wire = wire
        self.matrix = (1/sp.sqrt(2)) * sp.Matrix([
            [1,1],
            [1,-1],
        ])

class X(SingleQubitGate):
    def __init__(self, wire):
        super().__init__()
        self.wire = wire
        self.matrix = sp.Matrix([
            [0,1],
            [1,0],
        ])

class Y(SingleQubitGate):
    def __init__(self, wire):
        super().__init__()
        self.wire = wire
        self.matrix = sp.Matrix([
            [0,-sp.I],
            [sp.I,0],
        ])

class Z(SingleQubitGate):
    def __init__(self, wire):
        super().__init__()
        self.wire = wire
        self.matrix = sp.Matrix([
            [1,0],
            [0,-1],
        ])

class RX(SingleQubitGate):
    def __init__(self, theta, wire):
        super().__init__()
        self.wire = wire
        half  = sp.Rational(1, 2)
        self.matrix = sp.Matrix([
            [sp.cos(half*theta), -sp.I*sp.sin(half*theta)], 
            [-sp.I*sp.sin(half*theta), sp.cos(half*theta)], 
        ])

class RY(SingleQubitGate):
    def __init__(self, theta, wire):
        super().__init__()
        self.wire = wire
        half  = sp.Rational(1, 2)
        self.matrix = sp.Matrix([
            [sp.cos(half*theta), -sp.sin(half*theta)], 
            [sp.sin(half*theta), sp.cos(half*theta)], 
        ])

class RZ(SingleQubitGate):
    def __init__(self, theta, wire):
        super().__init__()
        self.wire = wire
        half  = sp.Rational(1, 2)
        self.matrix = sp.Matrix([
            [sp.exp(-sp.I*half*theta), 0], 
            [0, sp.exp(sp.I*half*theta)], 
        ])

class RXX(TwoQubitGate):
    def __init__(self, theta, wire1, wire2):
        super().__init__()
        self.wire1 = wire1
        self.wire2 = wire2
        half  = sp.Rational(1, 2)
        self.matrix = sp.Matrix([
            [sp.cos(half*theta),0,0,-sp.I*sp.sin(half*theta)],
            [0,sp.cos(half*theta),-sp.I*sp.sin(half*theta),0],
            [0,-sp.I*sp.sin(half*theta),sp.cos(half*theta),0],
            [-sp.I*sp.sin(half*theta),0,0,sp.cos(half*theta)],
        ])
        self.submatrix = self.get_submatrix(wire1, wire2)

class RYY(TwoQubitGate):
    def __init__(self, theta, wire1, wire2):
        super().__init__()
        self.wire1 = wire1
        self.wire2 = wire2
        half  = sp.Rational(1, 2)
        self.matrix = sp.Matrix([
            [sp.cos(half*theta),0,0,sp.I*sp.sin(half*theta)],
            [0,sp.cos(half*theta),-sp.I*sp.sin(half*theta),0],
            [0,-sp.I*sp.sin(half*theta),sp.cos(half*theta),0],
            [sp.I*sp.sin(half*theta),0,0,sp.cos(half*theta)],
        ])
        self.submatrix = self.get_submatrix(wire1, wire2)

class RZZ(TwoQubitGate):
    def __init__(self, theta, wire1, wire2):
        super().__init__()
        self.wire1 = wire1
        self.wire2 = wire2
        half  = sp.Rational(1, 2)
        self.matrix = sp.Matrix([
            [sp.exp(-sp.I*half*theta),0,0,0],
            [0,sp.exp(sp.I*half*theta),0,0],
            [0,0,sp.exp(sp.I*half*theta),0],
            [0,0,0,sp.exp(-sp.I*half*theta)],
        ])
        self.submatrix = self.get_submatrix(wire1, wire2)

class SWAP(TwoQubitGate):
    def __init__(self, wire1, wire2):
        super().__init__()
        self.wire1 = wire1
        self.wire2 = wire2
        self.matrix = sp.Matrix([
            [1,0,0,0],
            [0,0,1,0],
            [0,1,0,0],
            [0,0,0,1],
        ])
        self.submatrix = self.get_submatrix(wire1, wire2)

class CX(TwoQubitGate):
    def __init__(self, wire1, wire2):
        super().__init__()
        self.wire1 = wire1
        self.wire2 = wire2
        self.matrix = sp.Matrix([
            [1,0,0,0],
            [0,1,0,0],
            [0,0,0,1],
            [0,0,1,0],
        ])
        self.submatrix = self.get_submatrix(wire1, wire2)

class CZ(TwoQubitGate):
    def __init__(self, wire1, wire2):
        super().__init__()
        self.wire1 = wire1
        self.wire2 = wire2
        self.matrix = sp.Matrix([
            [1,0,0,0],
            [0,1,0,0],
            [0,0,1,0],
            [0,0,0,-1],
        ])
        self.submatrix = self.get_submatrix(wire1, wire2)


import sympy as sp
import qiskit
from . import sp_gate
from . import sp_func

class Circuit:
    def __init__(self, num_qubits):
        self.num_qubits     = num_qubits
        self.qiskit_circuit = qiskit.circuit.QuantumCircuit(num_qubits)
        self.initial_state  = sp.Matrix([1]+[0]*(2**num_qubits-1))
        self.final_state    = None
        self.gate_list      = []
        self.symbol_set    = set()
        # for quantum machine learning
        self.meas_mode  = None
        self.meas_basis = None
    def evolve_state(self):
        self.final_state = self.initial_state
        gate_flag, gate_buffer = False, [sp_func.identity(2) for _ in range(self.num_qubits)]
        for i in range(len(self.gate_list)):
            if self.gate_list[i].gate_type == "SingleQubitGate":
                gate_flag = True
                wire = self.gate_list[i].wire
                gate_buffer[wire] = self.gate_list[i].matrix * gate_buffer[wire]
                if i == len(self.gate_list)-1:
                    self.final_state = sp_func.kron(*gate_buffer) * self.final_state
            elif self.gate_list[i].gate_type == "TwoQubitGate":
                if gate_flag == True:
                    self.final_state = sp_func.kron(*gate_buffer) * self.final_state
                    gate_flag, gate_buffer = False, [sp_func.identity(2) for _ in range(self.num_qubits)]
                submatrix = self.gate_list[i].submatrix
                wire1, wire2 = self.gate_list[i].wire1, self.gate_list[i].wire2
                wire_min, wire_max = min(wire1, wire2), max(wire1, wire2)
                full_matrix = sp_func.kron(sp_func.identity(2**wire_min), submatrix, sp_func.identity(2**(self.num_qubits-wire_max-1)))
                self.final_state = full_matrix * self.final_state
            else:
                raise TypeError("Unknown gate type : {self.gate_list[i].gate_type}")
    def measure(self, qubit, basis):
        assert self.final_state != None, "Circuit should be evolved before measurement: try Circuit.evolve_state()"
        assert basis in ["X", "Y", "Z"], f"Basis should be X or Y or Z, not {basis}"
        basis_dict = {
            "X": sp.Matrix([[0,1],[1,0],]),
            "Y": sp.Matrix([[0,-sp.I],[sp.I,0],]),
            "Z": sp.Matrix([[1,0],[0,-1],]),
        }
        full_matrix = sp_func.kron(sp_func.identity(2**qubit), basis_dict[basis], sp_func.identity(2**(self.num_qubits-qubit-1)))
        return (sp_func.adjoint(self.final_state) * full_matrix * self.final_state)[0]
    def draw(self, *args, **kwargs):
        return self.qiskit_circuit.draw(*args, **kwargs)
    def _check_wire(self, wire):
        num_qubits = self.num_qubits
        if wire >= num_qubits:
            raise ValueError(f"Wire {wire} is not between 0 to {num_qubits-1} with only {num_qubits} qubit(s).")
    def __add__(self, other):
        assert self.num_qubits >= other.num_qubits, "Composing small circuit with big circuit is not allowed."
        new_circuit = Circuit(self.num_qubits)
        new_circuit.qiskit_circuit = self.qiskit_circuit.compose(other.qiskit_circuit)
        new_circuit.gate_list      = self.gate_list + other.gate_list
        new_circuit.symbol_set     = self.symbol_set.union(other.symbol_set)
        if self.meas_mode != None: new_circuit.meas_mode = self.meas_mode
        elif other.meas_mode != None: new_circuit.meas_mode = other.meas_mode
        if self.meas_basis != None: new_circuit.meas_basis = self.meas_basis
        elif other.meas_basis != None: new_circuit.meas_basis = other.meas_basis
        return new_circuit
    def __call__(self, x):
        subs = {}
        for i in range(len(x)):
            subs[sp.Symbol(f"inputs_{i}", real=True)] = x[i]
        self.evolve_state()
        self.final_state = self.final_state.evalf(subs=subs)
        meas_value = []
        if self.meas_mode == "measure_all":
            for basis in self.meas_basis:
                meas_value += [self.measure(q, basis) for q in range(self.num_qubits)]
        elif self.meas_mode == "measure_single":
            for basis in self.meas_basis:
                meas_value += [self.measure(0, basis)]
        return sp.Matrix(meas_value)
    def h(self, wire):
        self._check_wire(wire)
        self.qiskit_circuit.h(wire)
        self.gate_list.append(sp_gate.H(wire))
    def x(self, wire):
        self._check_wire(wire)
        self.qiskit_circuit.x(wire)
        self.gate_list.append(sp_gate.X(wire))
    def y(self, wire):
        self._check_wire(wire)
        self.qiskit_circuit.y(wire)
        self.gate_list.append(sp_gate.Y(wire))
    def z(self, wire):
        self._check_wire(wire)
        self.qiskit_circuit.z(wire)
        self.gate_list.append(sp_gate.Z(wire))
    def rx(self, theta, wire):
        self._check_wire(wire)
        if type(theta) == str:
            theta = sp.Symbol(theta)
        if type(theta) == sp.Symbol:
            self.qiskit_circuit.rx(qiskit.circuit.Parameter(theta.name), wire)
            self.symbol_set.add(theta)
        else:
            self.qiskit_circuit.rx(theta, wire)
        self.gate_list.append(sp_gate.RX(theta, wire))
    def ry(self, theta, wire):
        self._check_wire(wire)
        if type(theta) == str:
            theta = sp.Symbol(theta)
        if type(theta) == sp.Symbol:
            self.qiskit_circuit.ry(qiskit.circuit.Parameter(theta.name), wire)
            self.symbol_set.add(theta)
        else:
            self.qiskit_circuit.ry(theta, wire)
        self.gate_list.append(sp_gate.RY(theta, wire))
    def rz(self, theta, wire):
        self._check_wire(wire)
        if type(theta) == str:
            theta = sp.Symbol(theta)
        if type(theta) == sp.Symbol:
            self.qiskit_circuit.rz(qiskit.circuit.Parameter(theta.name), wire)
            self.symbol_set.add(theta)
        else:
            self.qiskit_circuit.rz(theta, wire)
        self.gate_list.append(sp_gate.RZ(theta, wire))
    def rxx(self, theta, wire1, wire2):
        self._check_wire(wire1)
        self._check_wire(wire2)
        if type(theta) == str:
            theta = sp.Symbol(theta)
        if type(theta) == sp.Symbol:
            self.qiskit_circuit.rxx(qiskit.circuit.Parameter(theta.name), wire1, wire2)
            self.symbol_set.add(theta)
        else:
            self.qiskit_circuit.rxx(theta, wire1, wire2)
        self.gate_list.append(sp_gate.RXX(theta, wire1, wire2))
    def ryy(self, theta, wire1, wire2):
        self._check_wire(wire1)
        self._check_wire(wire2)
        if type(theta) == str:
            theta = sp.Symbol(theta)
        if type(theta) == sp.Symbol:
            self.qiskit_circuit.ryy(qiskit.circuit.Parameter(theta.name), wire1, wire2)
            self.symbol_set.add(theta)
        else:
            self.qiskit_circuit.ryy(theta, wire1, wire2)
        self.gate_list.append(sp_gate.RYY(theta, wire1, wire2))
    def rzz(self, theta, wire1, wire2):
        self._check_wire(wire1)
        self._check_wire(wire2)
        if type(theta) == str:
            theta = sp.Symbol(theta)
        if type(theta) == sp.Symbol:
            self.qiskit_circuit.rzz(qiskit.circuit.Parameter(theta.name), wire1, wire2)
            self.symbol_set.add(theta)
        else:
            self.qiskit_circuit.rzz(theta, wire1, wire2)
        self.gate_list.append(sp_gate.RZZ(theta, wire1, wire2))
    def swap(self, wire1, wire2):
        self._check_wire(wire1)
        self._check_wire(wire2)
        self.qiskit_circuit.swap(wire1, wire2)
        self.gate_list.append(sp_gate.SWAP(wire1, wire2))
    def cx(self, wire1, wire2):
        self._check_wire(wire1)
        self._check_wire(wire2)
        self.qiskit_circuit.cx(wire1, wire2)
        self.gate_list.append(sp_gate.CX(wire1, wire2))
    def cz(self, wire1, wire2):
        self._check_wire(wire1)
        self._check_wire(wire2)
        self.qiskit_circuit.cz(wire1, wire2)
        self.gate_list.append(sp_gate.CZ(wire1, wire2))


import sympy as sp

class Linear:
    def __init__(self, input_dim, output_dim, prefix="L"):
        self.input_dim  = input_dim
        self.output_dim = output_dim
        self.weights = [sp.Matrix(sp.symbols(f"{prefix}^{i+1}_0:{input_dim+1}", real=True)) for i in range(output_dim)]
    def __call__(self, x):
        assert len(x) == self.input_dim, f"input dimension not matched: Input={len(x)} | Expected={self.input_dim}"
        x = x.reshape(self.input_dim, 1)
        result = sp.zeros(self.output_dim, 1)
        for neuron in range(self.output_dim):
            weight = sp.Matrix(self.weights[neuron][1:])
            bias   = self.weights[neuron][0]
            result[neuron] += (weight.T * x)[0] + bias
        return result


import sympy as sp
from ..circuit.sp_circuit import Circuit

class Measurement(Circuit):
    def __init__(self, num_qubits, meas_mode, meas_basis):
        super().__init__(num_qubits)
        self.meas_mode = meas_mode
        self.meas_basis = meas_basis

class AngleEncoding(Circuit):
    def __init__(self, num_qubits, rot_gate="ry"):
        super().__init__(num_qubits)
        for i in range(num_qubits):
            getattr(self, rot_gate)(sp.Symbol(f"inputs_{i}", real=True), i)

class SingleRot(Circuit):
    def __init__(self, num_qubits, num_layers=1, prefix="T", 
                 rot_mode=['rz', 'ry', 'rz'], ent_mode='cx'):
        super().__init__(num_qubits)
        for l in range(num_layers):
            for q in range(num_qubits):
                for r in range(len(rot_mode)):
                    theta = sp.Symbol(prefix + f"^{l}_{q},{r}", real=True)
                    getattr(self, rot_mode[r])(theta, q)
            for q in range(num_qubits-1):
                getattr(self, ent_mode)(q, q+1)
            self.qiskit_circuit.barrier()


from setuptools import setup, find_packages

setup(
    name='qympy',
    version='1.0',
    license='MIT',
    author="Yi-An Chen",
    author_email='r08222011@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/r08222011/Qympy',
    keywords='quantum machine learning, quantum computation, sympy, qiskit',
    install_requires=[
        'numpy',
        'sympy',
        'qiskit',
        'pylatexenc',
        'matplotlib',
    ],
)